# Generated by Django 3.1.4 on 2020-12-12 18:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0006_datafile_backup_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='template',
            field=models.BooleanField(blank=True, default=False),
        ),
    ]
