# Generated by Django 3.1.4 on 2021-06-30 10:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0008_taskentry_metadata'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sequence',
            name='name',
            field=models.CharField(max_length=256),
        ),
    ]
