# Generated by Django 3.1.4 on 2023-11-08 15:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0011_auto_20211017_1548'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='rig_name',
            field=models.TextField(default='pagaiisland2'),
        ),
    ]
