# Generated by Django 2.2.1 on 2020-10-14 08:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0005_taskentry_sw_version'),
    ]

    operations = [
        migrations.AddField(
            model_name='datafile',
            name='backup_status',
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
    ]
