# Generated by Django 2.2.1 on 2019-12-02 08:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0003_auto_20191113_2158'),
    ]

    operations = [
        migrations.CreateModel(
            name='KeyValueStore',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.TextField()),
                ('value', models.TextField()),
            ],
        ),
    ]
